module GarlandRails
  class Base < Garland
    self.table_name = "garlands"

    validates :entity, presence: true
    validates_inclusion_of :entity_type, in: [DIFF, SNAPSHOT]

    # for given parent object, described by "belongs_to_id" and "belongs_to_type",
    # there are only one record of each type which is head or tail
    validates_uniqueness_of :type, scope: [
      :belongs_to_id,
      :belongs_to_type,
      :next_id,
    ], conditions: -> { where("next_id IS NULL OR previous_id IS NULL") }

    # Minitest raises "No visible difference in the *#inspect output" error
    # while comparing Garland objects without this.
    def ==(other)
      self.attributes == other.attributes
    end

    # everything seems to work without `polymorphic: true`,
    # but we will set it just for accordance to docs
    def self.belongs_to(name, scope = nil, options = {}, &extension)
      if scope.class == Hash
        options = scope
        scope = nil
      end
      options = options.merge(foreign_key: "belongs_to_id")

      super(name, scope, options, &extension)
    end
  end
end
