module GarlandRails
  class Base < Garland
    def self.push(args)
      return nil unless args.class == Hash

      if args[:hash]
        hash = args[:hash]
        return nil unless hash.class == Hash

        belongs_to = args[:belongs_to]
        if belongs_to
          belongs_to_params = self._split_belongs_to(belongs_to)
          belongs_to_id, belongs_to_type =
            belongs_to_params.values_at(:belongs_to_id, :belongs_to_type)
        end
      else
        hash = args
        belongs_to = nil
        belongs_to_id = nil
        belongs_to_type = nil
      end

      head = self.head(belongs_to)
      if head
        diff = self.insert_diff(hash, belongs_to)
      else
        diff = self.init(hash, belongs_to)
      end

      diff
    end

    def self.init(hash, belongs_to = nil)
      common_props = self._split_belongs_to(belongs_to)

      tail_props = common_props.merge(entity: {}.to_s, entity_type: SNAPSHOT)
      brand_new_tail = self.new(tail_props)

      diff = HashDiffSym.diff({}, hash)
      first_diff_props = common_props.merge(entity: diff.to_s, entity_type: DIFF)
      first_diff = self.new(first_diff_props)

      head_props = common_props.merge(entity: hash.to_s, entity_type: SNAPSHOT)
      brand_new_head = self.new(head_props)

      self.transaction do
         ActiveRecord::Base.connection.create_savepoint("savepoint_before_init")

         # first id: tail ({})
         # second id: head (latest snapshot)
         # third+: diffs
         unless brand_new_tail.save
           Rails.logger.error("Unable to create new tail with props '#{tail_props}'")
           ActiveRecord::Base.connection.release_savepoint("savepoint_before_init")
           return nil
         end

         # belongs_to validations were in `brand_new_tail.save`
         # here and below validations may be skipped as long as we check for continuity later
         brand_new_head.save(validate: false)
         first_diff.save(validate: false)
         brand_new_tail.update_attribute(:next_id, first_diff.id)
         first_diff.update_attribute(:previous_id, brand_new_tail.id)
         first_diff.update_attribute(:next_id, brand_new_head.id)
         brand_new_head.update_attribute(:previous_id, first_diff.id)

         unless self.continuous?(belongs_to)
           Rails.logger.error("Initialized garland is not continuous")
           ActiveRecord::Base.connection.exec_rollback_to_savepoint("savepoint_before_init")
           ActiveRecord::Base.connection.release_savepoint("savepoint_before_init")
           return nil
         end
      end

      first_diff
    end

    def self.insert_diff(hash, belongs_to = nil)
      head = self.head(belongs_to)
      last_diff = self.find_by(id: head.previous_id)
      common_props = self._split_belongs_to(belongs_to)

      diff = HashDiffSym.diff(head.safe_eval_entity, hash)
      return unless diff.any?

      new_diff_props = common_props.merge(
        entity: diff.to_s,
        entity_type: DIFF,
        previous_id: last_diff.id,
        next_id: head.id,
      )
      new_diff = self.new(new_diff_props)

      self.transaction do
         ActiveRecord::Base.connection.create_savepoint("savepoint_before_insert_diff")

         # insert_diff should not use skipping valudatuons methods
         # because we don't want to check for continuity on every push
         unless new_diff.save
           Rails.logger.error("Unable to create new_diff with props '#{new_diff_props}'")
           ActiveRecord::Base.connection.release_savepoint("savepoint_before_init")
           return nil
         end

         last_diff.next_id = new_diff.id
         unless last_diff.save
           Rails.logger.error("Unable to save last_diff with 'next_id' = '#{new_diff.id}'")
           ActiveRecord::Base.connection.exec_rollback_to_savepoint("savepoint_before_insert_diff")
           ActiveRecord::Base.connection.release_savepoint("savepoint_before_init")
           return nil
         end

         head.previous_id = new_diff.id
         unless head.save
           Rails.logger.error("Unable to save head with 'previous_id' = '#{new_diff.id}'")
           ActiveRecord::Base.connection.exec_rollback_to_savepoint("savepoint_before_insert_diff")
           ActiveRecord::Base.connection.release_savepoint("savepoint_before_init")
           return nil
         end

         head.entity = hash.to_s
         unless head.save
           Rails.logger.error("Unable to save head with 'entity' = '#{hash}'")
           ActiveRecord::Base.connection.exec_rollback_to_savepoint("savepoint_before_insert_diff")
           ActiveRecord::Base.connection.release_savepoint("savepoint_before_init")
           return nil
         end
      end

      new_diff
    end
  end
end
