module GarlandRails
  class Base < Garland
    # REVIEW: shouldn't thread(), tail() and head() have the same args as push()?
    def self.thread(belongs_to = nil)
      if belongs_to
        return self.where(
          "belongs_to_id = ? AND belongs_to_type = ?",
          belongs_to.id, table_type(belongs_to)
        )
      else
        return self.where("belongs_to_id is null AND belongs_to_type is null")
      end
    end

    def self.tail(belongs_to = nil)
      self.thread(belongs_to).find_by(previous_id: nil)
    end

    def self.head(belongs_to = nil)
      self.thread(belongs_to).find_by(next_id: nil)
    end

    def self.last_diff(belongs_to = nil)
      head = self.head(belongs_to)
      return nil unless head

      head.previous
    end

    def self.continuous?(belongs_to = nil)
      tail = self.tail(belongs_to)
      head = self.head(belongs_to)
      return false unless tail && head

      current_bulb = tail
      current_hash = tail.safe_eval_entity
      items_counted = 1
      while current_bulb.next_id do
        items_counted += 1
        current_bulb = current_bulb.next
        if current_bulb.entity_type == DIFF
          current_hash = HashDiffSym.patch!(current_hash, current_bulb.safe_eval_entity)
        else
          break
        end
      end

      items_counted == self.thread(belongs_to).size && current_hash == head.safe_eval_entity
    end
    
    def self.table_type(record)
      record.class.name
    end

    def self.any?(belongs_to = nil)
      self.thread(belongs_to).any?
    end

    def next
      Garland.find(self.next_id) if self.next_id
    end

    def previous
      Garland.find(self.previous_id) if self.previous_id
    end

    def safe_eval_entity
      return nil unless self.entity =~ /\[.*\]/ || self.entity =~ /{.*}/

      eval(self.entity)
    end

    private
      def self._split_belongs_to(belongs_to)
        if belongs_to
          belongs_to_id = belongs_to.id
          belongs_to_type = table_type(belongs_to)
        else
          belongs_to_id = nil
          belongs_to_type = nil
        end

        { belongs_to_id: belongs_to_id, belongs_to_type: belongs_to_type }
      end
  end
end
